# =========================================
# Main file for Unwrap and Flatten Faces
# Copyright 2014-2017, Alexander C. Schreyer
# =========================================


require 'sketchup.rb'


# =========================================


module AS_Extensions

  module AS_Flatten
  
  
  # =========================================

 
    # Get settings
    @conf = Sketchup.read_default @extname, "confirmation", "显示"
    @color = Sketchup.read_default @extname, "colorize", "否"
    @prompt = Sketchup.read_default @extname, "prompts", "隐藏"
    @iter = Sketchup.read_default @extname, "iterations", "1000"
    @axis = Sketchup.read_default @extname, "axis", "Z_AXIS"
    
    # Define some random colors and an index
    @c = []
    50.times {
      @c << Sketchup::Color.new(rand(70..230), rand(70..230), rand(70..230))
    }
    @c_i = 0


  # =========================================
  
  
    def self.rotate_drop( gr )
    # Rotates and drops a group of flattened faces to the ground (z=0)

      # Get first face in group
      fa = gr.entities.grep( Sketchup::Face )[0]
      norm = fa.normal.transform!( gr.transformation )
      
      UI.messagebox("现在旋转\n是否继续？") if @prompt == "显示"
            
      # Then rotate the group - if necessary - so that it lays flat
      if !norm.parallel?( Object.const_get(@axis) )
        rot = norm.angle_between( Object.const_get(@axis) )
        cr = norm.cross( Object.const_get(@axis) )
        t1 = Geom::Transformation.rotation( gr.bounds.center , cr , rot )
        gr.transform!( t1 )
      end
      
      UI.messagebox("现在将Z轴归零\n是否继续？") if @prompt == "显示"

      # Then drop to zero
      cen = gr.bounds.center
      case @axis
        when "X_AXIS"
          t2 = Geom::Transformation.translation( [-cen.x,0,0] )
        when "Y_AXIS"
          t2 = Geom::Transformation.translation( [0,-cen.y,0] )          
        else
          t2 = Geom::Transformation.translation( [0,0,-cen.z] )
      end
      gr.transform!( t2 )
    
    end  


  # =========================================
  
  
    def self.flatten_faces
    # Flattens (into x-y plane) all selected faces (without distortions!)

      mod = Sketchup.active_model
      ent = mod.entities
      sel = mod.selection

      # Arrays to hold the faces
      ofaces = Array.new

      # Make a collection of all faces in the selection
      ofaces = sel.grep( Sketchup::Face )
      startnum = ofaces.length

      # Reminder if started from menu item and nothing selected
      if ofaces.length < 1

        UI.messagebox("请先选择至少一个未群组或组件的面以使用此工具。")

      else # Get going

        # Base next steps on the first face
        e = ofaces[0]

        # Check if all selected faces are coplanar
        coplanar = true
        ofaces.each {|f|
          coplanar = false if not f.normal.parallel? e.normal
        }

        if coplanar  # Just group faces without unwrapping, then lay flat

          mod.start_operation("展平面")
          
          # Colorize original faces if desired
          ofaces.each{ |f| f.material = @c[@c_i] if f.is_a? Sketchup::Face } if @color == "是"

          # Group face(s) and drop
          g = ent.add_group()
          g2 = ent.add_group( ofaces )
          g3 = g.entities.add_instance( g2.definition , g.transformation.invert!*g2.transformation )
          g2.explode
          g3.explode  
          rotate_drop( g )
                      
          # Colorize flattened faces if desired
          g.entities.each{ |f| f.material = @c[@c_i] if f.is_a? Sketchup::Face } if @color == "是"
          @c_i = rand( @c.length )           

          # Allow for undo
          mod.commit_operation

        else  # Unwrap faces first, then lay them flat

          # Mention what we will do
          UI.messagebox("当前选择的为非共面的面。插件将先尝着试展开和压平，但并不是每次都能处理成功，如果遇到问题，可以试着选择较少的面来重新进行计算。由于每次计算都是随机的，因此结果可能会因有所不同。\n 此提示可以在设置里关闭！") if @conf == "显示"

          mod.start_operation("展开及压平")
          
          # Colorize original faces if desired
          ofaces.each{ |f| f.material = @c[@c_i] if f.is_a? Sketchup::Face } if @color == "是"

          # A few more variables for this
          faces = Array.new
          afaces = Array.new  # Need one more copy for iterations
          max_tries = @iter.to_i # Number of iterations to try
          n = 0

          # Find viable path through set of faces up to max. iterations
          until (faces.length == ofaces.length or n > max_tries-1)

            # Reset arrays for this iteration
            faces = []
            afaces = ofaces.dup  # use dup to make a duplicate here!
            n += 1

            # Give some feedback
            Sketchup.status_text = "展开 | 尝试迭代 #{n}"

            # Start with a random face in the set
            lface = afaces[rand(afaces.length-1)]
            faces << lface
            afaces.delete( lface )

            # Now walk through all the remaining faces and try to find a path
            for i in 1..afaces.length

              nface = nil

              # check all edges of a face
              lface.edges.each { |e|

                # only do this if two faces are at same edge
                if e.faces.length > 1
                  # get tentative new face
                  tface = (e.faces - [lface]).first
                  # check if it's still available
                  if (afaces.index(tface) != nil)
                    # add it to the array and use it as the new start face
                    nface = tface
                    afaces.delete( nface )
                    faces << nface
                    lface = nface
                  end
                end
                # we have found a face!
                break if (nface != nil)

              }

            end

          end

          # Now do the actual unwrapping
          
          begin # Catch problems while unwrapping

            # Group for all unwrapped faces
            g = ent.add_group

            # Now iterate through the faces
            (1..faces.length-1).each {|i|

              # Find common edge
              cedge = faces[i].edges & faces[i-1].edges

              # Get target normal and do the transformation math
              tnormal = faces[i].normal
              onormal = faces[i-1].normal
              rot = tnormal.angle_between( onormal )
              cr = tnormal.cross( onormal )
              # Can't do this if faces are coplanar, so add if
              t1 = Geom::Transformation.rotation( cedge[0].start , cr , -rot ) if rot != 0.0
              
              # Add the face to the done group             
              g2 = ent.add_group( faces[i-1] )
              g3 = g.entities.add_instance( g2.definition , g.transformation.invert!*g2.transformation )              
              g2.explode
              g3.explode             
              
              # Then rotate complete group to next face  
              g.transform!( t1 ) if rot != 0.0
              
              UI.messagebox( "面 #{i} 完成\n是否继续？" ) if @prompt == "显示"

            }
            
            # Add the last face to the done group             
            g2 = ent.add_group( faces[faces.length-1] )
            g3 = g.entities.add_instance( g2.definition , g.transformation.invert!*g2.transformation )
            g2.explode
            g3.explode    
            UI.messagebox( "面 #{faces.length} 完成\n是否继续？" ) if @prompt == "显示"

            # Rotate after unwrapping and drop   
            endnum = g.entities.grep( Sketchup::Face ).length
            rotate_drop( g )
                            
            # Colorize flattened faces if desired
            g.entities.each{ |f| f.material = @c[@c_i] if f.is_a? Sketchup::Face } if @color == "是"
            @c_i = rand( @c.length ) 
            
            msg = "#{endnum} 从 #{startnum} 中选择展开的面。 "
            msg += "无法自动展开。 请尝试选择较少的面（或增加设置中的迭代）后进行重试。" if endnum < startnum
            msg += "展开后会产生重叠的面。 请尝试选择较少的面（分段展开）以获得更好的结果。" if endnum > startnum
            msg += "\n\n如果面不在地面上，则爆炸结果并再次运行工具。"
            UI.messagebox(msg) if @conf == "显示" or endnum != startnum
            if endnum == startnum
              Sketchup.status_text = "展开 | 完成" 
            else
              Sketchup.status_text = "展开 | 问题处理完成" 
            end

          rescue Exception => e  
            
            p e.backtrace.inspect
            Sketchup.status_text = "展开 | 问题处理完成"
            UI.messagebox("自动展开及压平的问题 (使用 #{n} 次尝试). 请撤消并使用较少的面来重试。 \n\n错误: #{e}")

          end

          # Allow for undo
          mod.commit_operation

        end

      end # main routine

    end # def flatten_faces
    
    
    # =========================================


    def self.smash_faces
    # Smashes (projects) the selected faces onto a plane (with distortions!)

      mod = Sketchup.active_model
      ent = mod.entities
      sel = mod.selection

      # Array to hold the faces
      ofaces = Array.new

      # Make a collection of all faces in the selection
      ofaces = sel.grep( Sketchup::Face )    
      
      # Reminder if started from menu item and nothing selected
      if ofaces.length < 1

        UI.messagebox("请先选择至少一个未群组或组件的面以使用此工具。")

      else # Get going      
      
        Sketchup.status_text = "粉碎 | 工作中..."
        
        mod.start_operation("粉碎曲面")
        
        # Colorize original faces if desired
        ofaces.each{ |f| f.material = @c[@c_i] if f.is_a? Sketchup::Face } if @color == "是"
      
        # Group for result and plane
        g = ent.add_group
        plane = [ ORIGIN , Object.const_get(@axis) ]
        
        # Now smash them (by copying)
        ofaces.each { |f|
          begin  # Catch double vertices error
            g.entities.add_face( f.vertices.map{ |i| i.position.project_to_plane( plane ) } )
          rescue
            p "忽略双顶点"
          end
        }      

        # Reverse any face's direction if needed
        faces = g.entities.grep( Sketchup::Face ) 
        faces.each { |f| f.reverse! if !f.normal.samedirection?( Object.const_get(@axis) ) }
        
        # Colorize flattened faces if desired
        g.entities.each{ |f| f.material = @c[@c_i] if f.is_a? Sketchup::Face } if @color == "是"
        @c_i = rand( @c.length ) 
        
        # Allow for undo
        mod.commit_operation
        
        Sketchup.status_text = "粉碎 | 完成"
        
      end

    end # smash_faces   


    # =========================================


    def self.contains_face
    # Checks if there is at least one face in the selection set

      contains = false
      Sketchup.active_model.selection.each{|e|
        contains = true if e.is_a? Sketchup::Face
      }
      return contains

    end # contains_face


    # =========================================


    def self.show_help
    # Show the website as an About dialog

      dlg = UI::WebDialog.new("#{@exttitle} - Help", true,'AS_Flatten_Help', 1100, 800, 150, 150, true)
      dlg.set_url('http://www.alexschreyer.net/projects/flatten-faces-plugin/')
      dlg.show

    end # show_help


    def self.show_zhcn
      dlg = UI::WebDialog.new("#{@exttitle} - 汉化网站", true,'AS_Flatten_Help', 1100, 800, 150, 150, true)
      dlg.set_url('https://www.liutao91.cn/design/sketchup/sketchup-plugin')
      dlg.show
    end # show_zhcn


    # =========================================
    
    
    def self.settings
    
      prompts = ["压平轴向" , "随机上色 ", "迭代次数 " , "确认对话框 " , "步骤提示 "]
      defaults = [@axis , @color, @iter , @conf , @prompt]
      lists = ["X_AXIS|Y_AXIS|Z_AXIS" , "是|否", "10|50|100|500|1000|5000|10000" , "显示|隐藏" , "显示|隐藏"]
      res = UI.inputbox( prompts , defaults , lists , "#{@exttitle} - 设置")
      if res
        Sketchup.write_default @extname, "axis", @axis = res[0]
        Sketchup.write_default @extname, "colorize", @color = res[1]
        Sketchup.write_default @extname, "iterations", @iter = res[2].to_i
        Sketchup.write_default @extname, "confirmation", @conf = res[3]  
        Sketchup.write_default @extname, "prompts", @prompt = res[4]
      end
      
    end
    
    
    # =========================================


    # Load plugin at startup and add menu items to context menu
    if !file_loaded?(__FILE__)

      # Add to the tools menu
      tmenu = UI.menu("Tools").add_submenu( @exttitle )
      tmenu.add_item("展开及压平") { self.flatten_faces }
      tmenu.add_item("粉碎项目") { self.smash_faces }
      tmenu.add_item("设置") { self.settings }
      tmenu.add_item("帮助") { self.show_help }
      tmenu.add_item("汉化网站") { self.show_zhcn }
      # Add to the context menu
      UI.add_context_menu_handler do |menu|
        if( self.contains_face )
          sub = menu.add_submenu( @exttitle )
          sub.add_item("展开及压平") { self.flatten_faces }
          sub.add_item("粉碎项目") { self.smash_faces }
          sub.add_item("设置") { self.settings }
        end
      end

    # 以下部分是添加图标的代码
    toolbar = UI::Toolbar.new "展开及压平"
    cmd = UI::Command.new("展开及压平") { self.flatten_faces }
    cmd.tooltip = "展开及压平"
    cmd.status_bar_text = '展开及压平选定的面，请先选择至少一个未群组或组件的面以使用此工具。'
    cmd.small_icon = "flatten-icon-60.png"
    cmd.large_icon = "flatten-icon-60.png"
    toolbar = toolbar.add_item cmd
    toolbar.show
    puts cmd.large_icon
    # 以上部分是添加图标的代码  
      # Let Ruby know we have loaded this file
      file_loaded(__FILE__)

    end # if


    # =========================================


  end # module AS_Flatten

end # module AS_Extensions


# =========================================
